/*
 * This is the main driver for the Cyclades Cyclom-Z for FreeBSD 2.x and 3.x.
 * For more information, see the RELNOTES file
 */

/* Pull in the include file generated by config */
#include "cz.h"

/* Include all the standard include files */
#if NCZ > 0

#include <unistd.h>
#include <sys/param.h>
#include <sys/systm.h>
#include <sys/malloc.h>
#include <sys/reboot.h>
#include <sys/fcntl.h>
#include <sys/proc.h>
#include <sys/conf.h>
#include <sys/file.h>
#include <sys/uio.h>
#include <sys/tty.h>
#include <sys/kernel.h>
#include <sys/syslog.h>
#include <machine/clock.h>
#include <vm/vm.h>
#include <vm/pmap.h>
#include <pci/pcivar.h>


/* Get the DEVFS stuff */
#ifdef DEVFS
#include <sys/devfsext.h>
#endif				/* DEVFS */

/* Include driver-specific headers */

/* Hardware info */
#include <dev/cz/cyclomz.h>

/* Firmware info */
#include <dev/cz/zfwint.h>

/* The firmware image */
#include <dev/cz/zfirm.h>

/* Driver specific configs and stuff that don't live elsewhere */
#include <dev/cz/lcz.h>

/* Function prototypes for things we'll need later */
static char    *cz_probe __P((pcici_t, pcidi_t));
static void cz_attach __P((pcici_t, int));
#ifdef CZINT
static void cz_intr __P((void *));
#endif				/* CZINT */
static u_long   cz_count = 0;

static d_open_t cz_open;
static d_close_t cz_close;
static d_read_t cz_read;
static d_write_t cz_write;
static d_ioctl_t cz_ioctl;
static d_stop_t cz_stop;
static d_devtotty_t cz_devtotty;

/* Set up the board's device structure */
static struct pci_device cz_device = {"cz", cz_probe, cz_attach, &cz_count, NULL};
DATA_SET(pcidevice_set, cz_device);

/* Flag for whether we should initialize our structures on first-probe basis */
static unsigned char first_probe = 1;

/* Define our major number */
#define CDEV_MAJOR 48

/* define our cdevsw structure */
static struct cdevsw cz_cdevsw =
{
	cz_open,
	cz_close,
	cz_read,
	cz_write,
	cz_ioctl,
	cz_stop,
	noreset,
	cz_devtotty,
#if __FreeBSD_version < 300000
	ttselect,
#else
	ttpoll,
#endif
	nommap,
	NULL,
	"cz",
	NULL,
	CDEV_MAJOR,
};

/* Define the pointers for DEVFS (they're never used once devfs is up) */
#ifdef DEVFS

void           *devfs_token_ttyd;
void           *devfs_token_ttyl;
void           *devfs_token_ttyi;
void           *devfs_token_cuaa;
void           *devfs_token_cual;
void           *devfs_token_cuai;
#endif				/* DEVFS */


/* Routine for polling Cyclom-Z board */
#ifndef CZINT
static void
cz_poll(void *poll_board)
{
	struct board_info *board;

	struct tty     *tp;

	long unsigned int port;
	long unsigned int get, put, size;
	long unsigned int buffer_space, queue_space, moveable, moved, count, ttyqueue,
	                bqreturn = 0;

	unsigned char   temp_char;
	unsigned char  *buffer_overlay;
	board = (struct board_info *) poll_board;
	for (port = 0; (port < board->board_ctrl->n_channel); port++) {

		tp = &board->ports[port].ttyinfo;
		if (tp->t_state & TS_ISOPEN) {
			/* TX Code */
			buffer_overlay = (unsigned char *) (board->window + board->ports[port].buffer_ctrl->tx_bufaddr);
			put = board->ports[port].buffer_ctrl->tx_put;
			get = board->ports[port].buffer_ctrl->tx_get;
			size = board->ports[port].buffer_ctrl->tx_bufsize;
			buffer_space = tx_moveable(put, get, size);
			while ((buffer_space > 0) && (tp->t_outq.c_cc > 0)) {
				board->ports[port].flags |= DO_POLL_WAKEUP;
				moveable = MIN(buffer_space, tp->t_outq.c_cc);
				moved = q_to_b(&tp->t_outq, &buffer_overlay[put], moveable);
				if (moved != moveable)
					printf("cz:q_to_b error on transmit in poll routine\n");
				put = ((put + moved) % size);
				buffer_space = tx_moveable(put, get, size);
			}
			board->ports[port].buffer_ctrl->tx_put = put;

			if (tp->t_outq.c_cc > 0)
				tp->t_state |= TS_BUSY;
			else
				tp->t_state &= ~(TS_BUSY);
			/* RX Code */
			buffer_overlay = (unsigned char *) (board->window + board->ports[port].buffer_ctrl->rx_bufaddr);
			put = board->ports[port].buffer_ctrl->rx_put;
			get = board->ports[port].buffer_ctrl->rx_get;
			size = board->ports[port].buffer_ctrl->rx_bufsize;
			if (tp->t_state & TS_CAN_BYPASS_L_RINT) {
				buffer_space = rx_moveable(put, get, size);
				bqreturn = 0;
				while ((buffer_space > 0) && (bqreturn == 0)) {
				        board->ports[port].flags |= DO_POLL_WAKEUP;
					bqreturn = b_to_q(&buffer_overlay[get], buffer_space, &tp->t_rawq);
					get = ((get + (buffer_space - bqreturn)) % size);
					buffer_space = rx_moveable(put, get, size);

				}
			} else {
				ttyqueue = (TTYHOG - (tp->t_rawq.c_cc + tp->t_canq.c_cc));
				while ((ttyqueue > 0) && (get != put)) {
					board->ports[port].flags |= DO_POLL_WAKEUP;
					(*linesw[tp->t_line].l_rint) (buffer_overlay[get], tp);
					get = ((get + 1) % size);
					ttyqueue--;
				}
			}
			board->ports[port].buffer_ctrl->rx_get = get;
		}
		/* Code for CD handling */
		if ((board->ports[port].channel_ctrl->rs_status & C_RS_DCD) && ((tp->t_state & TS_CARR_ON) == 0)) {
			tp->t_state |= TS_CARR_ON;
			(*linesw[tp->t_line].l_modem) (tp, 1);
			board->ports[port].flags |= DO_POLL_WAKEUP;
		}
		if (((board->ports[port].channel_ctrl->rs_status & C_RS_DCD) == 0) && (tp->t_state & TS_CARR_ON)) {
			if ((board->ports[port].flags & PORT_CALL_OUT_ACTIVE) == 0) {
				tp->t_state &= ~(TS_CARR_ON);
				(*linesw[tp->t_line].l_modem) (tp, 0);
				board->ports[port].flags |= DO_POLL_WAKEUP;
			}
		}
		if (board->ports[port].flags & DO_POLL_WAKEUP) {
			wakeup(tp);
			ttwakeup(tp);
			ttwwakeup(tp);
			board->ports[port].flags &= ~(DO_POLL_WAKEUP);
		}
	}
	timeout(cz_poll, poll_board, hz / CZPOLLVAL);
	return;
}

#else				/* CZINT */

static unsigned char expected;
/* Code for interrupt handling */
static void
cz_intr(void *token)
{
	struct board_info *board;

	struct tty     *tp;

	long unsigned int interrupt_reason;
	long unsigned int port;
	volatile long unsigned int get, put, size;
	volatile long unsigned int buffer_space, moveable, moved, ttyqueue, count,
	                bqreturn;

	unsigned char  *buffer_overlay;

	board = (struct board_info *) token;
	port = board->board_ctrl->fwcmd_channel;
	interrupt_reason = board->plx9060->loc_doorbell;
	if (interrupt_reason == 0) {
		/* Stray interrupt? Or IRQ Sharing? */
		return;
	}
	while (interrupt_reason != 0) {
		board->plx9060->loc_doorbell = interrupt_reason;

		tp = &board->ports[port].ttyinfo;

		switch (interrupt_reason) {
		case C_CM_INTBACK:	/* May be used by cz_start() */
		case C_CM_TXBEMPTY:	/* Used by FW for TX buffer empty */
		case C_CM_TXLOWWM:	/* Used by FW for TX buffer low   */
			{
				buffer_overlay = (unsigned char *) (board->window + board->ports[port].buffer_ctrl->tx_bufaddr);
				put = board->ports[port].buffer_ctrl->tx_put;
				get = board->ports[port].buffer_ctrl->tx_get;
				size = board->ports[port].buffer_ctrl->tx_bufsize;
				buffer_space = tx_moveable(put, get, size);
				while ((buffer_space > 0) && (tp->t_outq.c_cc > 0)) {
					moveable = MIN(buffer_space, tp->t_outq.c_cc);
					moved = q_to_b(&tp->t_outq, &buffer_overlay[put], moveable);
					put = ((put + moved) % size);
					buffer_space = tx_moveable(put, get, size);
				}
				board->ports[port].buffer_ctrl->tx_put = put;
				if (tp->t_outq.c_cc > 0)
					tp->t_state |= TS_BUSY;
				else {
					tp->t_state &= ~(TS_BUSY);
					tp->t_state |= TS_SO_OCOMPLETE;
					tp->t_state |= TS_SO_OLOWAT;
				}
				/* break; */
			}
		case C_CM_RXHIWM:	/* Used by FW for RX Buffer near full      */
		case C_CM_RXNNDT:	/* Used by FW for no new data in RX
					 * buffer */
			{
				get = board->ports[port].buffer_ctrl->rx_get;
				put = board->ports[port].buffer_ctrl->rx_put;
				size = board->ports[port].buffer_ctrl->rx_bufsize;
				buffer_overlay = (unsigned char *) (board->window + board->ports[port].buffer_ctrl->rx_bufaddr);

				if ((tp->t_state & TS_ISOPEN) == 0)
					return;

				if (tp->t_state & TS_CAN_BYPASS_L_RINT) {
					buffer_space = rx_moveable(put, get, size);
					bqreturn = 0;
					while ((buffer_space > 0) && (bqreturn == 0)) {
						bqreturn = b_to_q(&buffer_overlay[get], buffer_space, &tp->t_rawq);
						get = ((get + (buffer_space - bqreturn)) % size);
						buffer_space = rx_moveable(put, get, size);
					}
				} else {
					while (get != put) {
						(*linesw[tp->t_line].l_rint) (buffer_overlay[get], tp);
						get = ((get + 1) % size);
					}
				}
				board->ports[port].buffer_ctrl->rx_get = get;
				break;
			}
		case C_CM_MDCD:
			{
				if (board->ports[port].channel_ctrl->rs_status & C_RS_DCD) {
					tp->t_state |= TS_CARR_ON;
					(*linesw[tp->t_line].l_modem) (tp, 1);
				} else {
					if (!(board->ports[port].flags & PORT_CALL_OUT_ACTIVE)) {
						(*linesw[tp->t_line].l_modem) (tp, 0);
						while (board->plx9060->pci_doorbell != 0);
						board->ports[port].channel_ctrl->rs_control &= ~(C_RS_DTR);
						board->plx9060->pci_doorbell = C_CM_IOCTLM;
						board->ports[port].flags |= DTR_FORCED_OFF;
						timeout(cz_dtr_on, (void *) &tp->t_dev, CZ_DTR_TICKS);
					}
					tp->t_state &= ~(TS_CARR_ON);
				}
				break;
			}
		case C_CM_IOCTLW:
			{
				return;
			}
		default:
			{
				break;
			}
		}
		wakeup(tp);
		ttwakeup(tp);
		ttwwakeup(tp);
		interrupt_reason = board->plx9060->loc_doorbell;
	}
	return;
}
#endif				/* CZINT */

/* Routine to reset dtr flag after call drop */
static void
clear_dtr_flag(void *devt)
{
	unsigned int    czminor, board, port = 0;
	czminor = minor(*(dev_t *) devt);
	board = CZBOARD(czminor);
	port = CZPORT(czminor);
	czcontrol.board[board].ports[port].flags &= ~DTR_FORCED_OFF;
	wakeup(&czcontrol.board[board].ports[port].flags);
}

/* Routine to reactivate DTR after a specified period */
static void
cz_dtr_on(void *devt)
{
	unsigned int    czminor, board, port = 0;
	czminor = minor(*(dev_t *) devt);
	board = CZBOARD(czminor);
	port = CZPORT(czminor);
	while (czcontrol.board[board].plx9060->pci_doorbell != 0);
	czcontrol.board[board].ports[port].channel_ctrl->rs_control |= C_RS_DTR;
	czcontrol.board[board].board_ctrl->hcmd_channel = port;
	czcontrol.board[board].plx9060->pci_doorbell = C_CM_IOCTLM;
	while (czcontrol.board[board].plx9060->pci_doorbell != 0);
	clear_dtr_flag(devt);
}
/* Routine to brand hardware to line settings */
void
cz_set_termios_params(unsigned int board, unsigned int port)
{
	struct tty     *tp;
	struct CH_CTRL *chctrl;
	tp = &czcontrol.board[board].ports[port].ttyinfo;
	chctrl = czcontrol.board[board].ports[port].channel_ctrl;
	while (czcontrol.board[board].plx9060->pci_doorbell != 0);
	tp->t_termios.c_ispeed = tp->t_termios.c_ospeed;
	switch (tp->t_termios.c_ospeed) {
	case B0:
		{
			break;
		}
	case B50:		/* Use this for 230400 w/apps that don't
				 * support it */
		{
			chctrl->comm_baud = 230400;
			break;
		}
	case B75:		/* Use this for 460800 */
		{
			chctrl->comm_baud = 460800;
			break;
		}
	case B150:
		{
			chctrl->comm_baud = 921600;
			break;
		}
	default:
		{
			chctrl->comm_baud = czcontrol.board[board].ports[port].ttyinfo.t_termios.c_ospeed;
			break;
		}
	}
	chctrl->comm_data_l = 0L;
	switch (tp->t_termios.c_cflag & CSIZE) {
	case CS5:
		{
			chctrl->comm_data_l |= C_DL_CS5;
			break;
		}
	case CS6:
		{
			chctrl->comm_data_l |= C_DL_CS6;
			break;
		}
	case CS7:
		{
			chctrl->comm_data_l |= C_DL_CS7;
			break;
		}
	case CS8:
		{
			chctrl->comm_data_l |= C_DL_CS8;
			break;
		}
	}
	if (tp->t_termios.c_cflag & CSTOPB) {
		if ((chctrl->comm_data_l & C_DL_CS) == C_DL_CS5) {
			chctrl->comm_data_l |= C_DL_15STOP;
		} else {
			chctrl->comm_data_l |= C_DL_2STOP;
		}
	} else {
		chctrl->comm_data_l |= C_DL_1STOP;
	}
	if (tp->t_termios.c_cflag & PARENB) {
		if (tp->t_termios.c_cflag & PARODD) {
			chctrl->comm_parity = C_PR_ODD;
		} else {
			chctrl->comm_parity = C_PR_EVEN;
		}
	} else {
		chctrl->comm_parity = C_PR_NONE;
	}

	/* RTS/CTS FLOW */
	if ((tp->t_termios.c_cflag & CRTSCTS) || (tp->t_termios.c_cflag & CCTS_OFLOW)) {
		chctrl->hw_flow |= (C_RS_CTS | C_RS_RTS);
	} else {
		chctrl->hw_flow &= ~(C_RS_CTS | C_RS_RTS);
	}
	/* DSR FLOW */
	if (tp->t_termios.c_cflag & CDSR_OFLOW) {
		chctrl->hw_flow |= C_RS_DSR;
	} else {
		chctrl->hw_flow &= ~(C_RS_DSR);
	}

	/* DTR FLOW */
	if (tp->t_termios.c_cflag & CDTR_IFLOW) {
		chctrl->hw_flow |= C_RS_DTR;
	} else {
		chctrl->hw_flow &= ~(C_RS_DTR);
	}

	/* CD FLOW */
	if (tp->t_termios.c_cflag & CCAR_OFLOW) {
		chctrl->hw_flow |= C_RS_DCD;
	} else {
		chctrl->hw_flow &= ~(C_RS_DCD);
	}
	while (czcontrol.board[board].plx9060->pci_doorbell != 0);
	czcontrol.board[board].board_ctrl->hcmd_channel = port;
	czcontrol.board[board].plx9060->pci_doorbell = C_CM_IOCTLW;
	while (czcontrol.board[board].plx9060->pci_doorbell != 0);
	czcontrol.board[board].board_ctrl->hcmd_channel = port;
	czcontrol.board[board].plx9060->pci_doorbell = C_CM_IOCTLM;
	return;
}


/* Routine to get data flowing to the board (hence triggering interrupts) */
void
cz_start(struct tty * tty)
{
	unsigned int    spl_value;
	unsigned long int czminor, board, port, type = 0;
	unsigned long int get, put, size;
	unsigned long int buffer_space, moveable, moved;
	struct tty     *tp;
	unsigned char  *buffer_overlay;
	unsigned char   temp_char;
	czminor = minor(tty->t_dev);
	board = CZBOARD(czminor);
	port = CZPORT(czminor);
	type = CZPORTTYPE(czminor);
	tp = &czcontrol.board[board].ports[port].ttyinfo;
	if (tp->t_outq.c_cc < 1)
		return;

	spl_value = spltty();
	if (tp->t_state & TS_BUSY) {
		splx(spl_value);
		return;
	}
	buffer_overlay = (unsigned char *) (czcontrol.board[board].window + czcontrol.board[board].ports[port].buffer_ctrl->tx_bufaddr);
	put = czcontrol.board[board].ports[port].buffer_ctrl->tx_put;
	get = czcontrol.board[board].ports[port].buffer_ctrl->tx_get;
	size = czcontrol.board[board].ports[port].buffer_ctrl->tx_bufsize;
	buffer_space = tx_moveable(put, get, size);
	while ((buffer_space > 0) && (tp->t_outq.c_cc > 0)) {
		moveable = MIN(buffer_space, tp->t_outq.c_cc);
		moved = q_to_b(&tp->t_outq, &buffer_overlay[put], moveable);
		if (moved != moveable)
			printf("cz: Error in interrupt TX q_to_b\n");
		put = ((put + moved) % size);
		buffer_space = tx_moveable(put, get, size);
	}
	czcontrol.board[board].ports[port].buffer_ctrl->tx_put = put;
	if (tp->t_outq.c_cc > 0)
		tp->t_state |= TS_BUSY;
	else {
		tp->t_state &= ~(TS_BUSY);
		tp->t_state |= TS_SO_OCOMPLETE;
		tp->t_state |= TS_SO_OLOWAT;
		ttwwakeup(tp);
	}
	splx(spl_value);
	return;
}

/* Routine to stop data flow */
static void
cz_stop(struct tty * tty, int value)
{
	unsigned int    czminor, board, port = 0;
	int             spl_value;
	if (value == 0)
		return;
	czminor = minor(tty->t_dev);
	board = CZBOARD(czminor);
	port = CZPORT(czminor);
	spl_value = spltty();
	if (value & FREAD)
		czcontrol.board[board].ports[port].buffer_ctrl->rx_get = czcontrol.board[board].ports[port].buffer_ctrl->rx_put;
	if (value & FWRITE)
		czcontrol.board[board].ports[port].buffer_ctrl->tx_put = czcontrol.board[board].ports[port].buffer_ctrl->tx_get;
	splx(spl_value);
	return;
}

/* Patch routine for the line discipline to cz_set_termios_params */
int
cz_param(struct tty * tty, struct termios * termios)
{
	unsigned int    board, port, czminor = 0;
	czminor = minor(tty->t_dev);
	board = CZBOARD(czminor);
	port = CZPORT(czminor);
	czcontrol.board[board].ports[port].ttyinfo.t_termios = *termios;
	cz_set_termios_params(board, port);
	return 0;
}

/* Cyclom PCI Probe routine */
static char    *
cz_probe(pcici_t config_id, pcidi_t device_id)
{
	if (first_probe == 1) {
		czcontrol.active_boards = 0;
		first_probe = 0;
	}
	if ((device_id & 0xffff) != 0x120e)
		return NULL;
	if ((device_id >> 16) == 0x200)
		return "Cyclades Cyclom-Z Adapter (<1MB)";
	if ((device_id >> 16) == 0x201)
		return "Cyclades Cyclom-Z Adapter (>1MB)";

	return NULL;
}

/* The attach routine. Its ugly. It needs work. But, it gets the card going */
void
cz_attach(pcici_t config_id, int unit)
{
#ifdef DEVFS
	unsigned char   devfsname[12];
	unsigned int    devfsminor;
#endif				/* DEVFS */
	struct ZFILE_HEADER *header;
	struct ZFILE_CONFIG *config;
	struct ZFILE_BLOCK *block;
	long unsigned int mailbox, function;
	int             board = 0;
	dev_t           dev;
	struct CUSTOM_REG *custom_regs;
	unsigned long  *plx_regs;
	unsigned long   loc_buf[12], reg;
	unsigned short  port;
	int             i, j;
	vm_offset_t     vaddr, paddr;
#ifdef CZDEBUGFW
	function = 1;
#else
	function = 0;
#endif				/* ; */
	if ((czcontrol.active_boards + 1) > NCZ) {
		printf("cz%d: There are more boards then are configured. Change\n", czcontrol.active_boards);
		printf("cz%d: your kernel config file to contain more device\n", czcontrol.active_boards);
		printf("cz%d: statements for the cz driver\n", czcontrol.active_boards);
		return;
	}
	czcontrol.board[czcontrol.active_boards].config_id = config_id;
	if (pci_map_mem(config_id, CZ_PCI_PLX9060_MEM, &vaddr, &paddr) == 0) {
		printf("Could't map PLX9060 memory window\n");
		return;
	}
	czcontrol.board[czcontrol.active_boards].plx9060 = (struct RUNTIME_9060 *) vaddr;
	mailbox = czcontrol.board[czcontrol.active_boards].plx9060->mail_box_0;
	if (pci_map_port(config_id, CZ_PCI_PLX9060_PORT, &port) == 0) {
		printf("Couldn't map PLX9060 port window\n");
		return;
	}
	if (pci_map_mem(config_id, CZ_PCI_MEMORY_WINDOW, &vaddr, &paddr) == 0) {
		printf("Couldn't map Cyclom-Zo shared memory area\n");
		return;
	}
	czcontrol.board[czcontrol.active_boards].window = (unsigned char *) vaddr;

#ifdef CZINT
	if (pci_map_int(config_id, cz_intr, &czcontrol.board[czcontrol.active_boards], &tty_imask) == 0) {
		printf("Couldn't map interrupt\n");
		return;
	}
#endif				/* CZINT */
	czcontrol.board[czcontrol.active_boards].plx9060->loc_addr_base = WIN_RAM;

	plx_regs = (unsigned long *) czcontrol.board[czcontrol.active_boards].plx9060;
	for (i = 0; i < 12; i++)
		loc_buf[i] = *(plx_regs + i);
	reg = czcontrol.board[czcontrol.active_boards].plx9060->init_ctrl;
	reg |= 0x40000000;
	czcontrol.board[czcontrol.active_boards].plx9060->init_ctrl = reg;
	DELAY(250000);
	reg &= ~(0x40000000);
	czcontrol.board[czcontrol.active_boards].plx9060->init_ctrl = reg;
	DELAY(250000);
	for (i = 0; i < 12; i++)
		*(plx_regs + i) = loc_buf[i];

	/* Stop CPU */
	czcontrol.board[czcontrol.active_boards].plx9060->loc_addr_base = WIN_CREG;
	custom_regs = (struct CUSTOM_REG *) czcontrol.board[czcontrol.active_boards].window;
	custom_regs->cpu_stop = 0x00000000;

	header = (struct ZFILE_HEADER *) & z_fw[0];

	config = (struct ZFILE_CONFIG *) (&z_fw[0] + header->config_offset);

	i = 0;
	while ((i < header->n_config) && ((config->mailbox != mailbox) || (config->function != function)))
		config++;
	if (i >= header->n_config) {
		printf("Couldn't find an appropriate configuration\n");
		return;
	}
	if (mailbox > 0) {
		for (i = 0; i < config->n_blocks; i++) {
			block = (struct ZFILE_BLOCK *) (z_fw + header->block_offset + (sizeof(struct ZFILE_BLOCK) * config->block_list[i]));
			if (block->type == 1) {
				for (j = 0; j < block->size; j++) {
					*(czcontrol.board[czcontrol.active_boards].window) = z_fw[block->file_offset + j];
					DELAY(10);
				}
			}
		}
	}
	/* Load Firmware */
	czcontrol.board[czcontrol.active_boards].plx9060->loc_addr_base = WIN_RAM;
	for (i = 0; i < config->n_blocks; i++) {
		block = (struct ZFILE_BLOCK *) (z_fw + header->block_offset + (sizeof(struct ZFILE_BLOCK) * config->block_list[i]));
		if (block->type == 0) {
			for (j = 0; j < block->size; j++) {
				*(czcontrol.board[czcontrol.active_boards].window + block->ram_offset + j) = z_fw[block->file_offset + j];
			}

		}
	}
	/* Restart board */
	czcontrol.board[czcontrol.active_boards].plx9060->loc_addr_base = WIN_CREG;
	custom_regs->cpu_start = 0x00000000;
	czcontrol.board[czcontrol.active_boards].plx9060->loc_addr_base = WIN_RAM;
	czcontrol.board[czcontrol.active_boards].firmid = (struct FIRM_ID *) (czcontrol.board[czcontrol.active_boards].window + ID_ADDRESS);

	for (i = 0; i < 4; i++) {
		if (czcontrol.board[czcontrol.active_boards].firmid->signature != ZFIRM_ID)
			DELAY(250000);
		else
			break;
	}
	if (czcontrol.board[czcontrol.active_boards].firmid->signature != ZFIRM_ID) {
		if (czcontrol.board[czcontrol.active_boards].firmid->signature == ZFIRM_HLT) {
			printf("cz%d:Board was halted due to insufficient power on external\n", czcontrol.active_boards);
			printf("cz%d:expansion modules. Correctly power, and reboot\n", czcontrol.active_boards);
		} else {
			printf("Board not initialized.\n");
			printf("Was looking for %d. Got %d\n", ZFIRM_ID, czcontrol.board[czcontrol.active_boards].firmid->signature);
			printf("FPGA ID: %d  FPGA Version: %d\n", custom_regs->fpga_id, custom_regs->fpga_version);
		}
		return;
	}
	czcontrol.board[czcontrol.active_boards].zfw_ctrl = (struct ZFW_CTRL *) ((void *) (czcontrol.board[czcontrol.active_boards].window + czcontrol.board[czcontrol.active_boards].firmid->zfwctrl_addr));
	czcontrol.board[czcontrol.active_boards].board_ctrl = &czcontrol.board[czcontrol.active_boards].zfw_ctrl->board_ctrl;
	czcontrol.board[czcontrol.active_boards].board_ctrl->op_system = 0;
	czcontrol.board[czcontrol.active_boards].board_ctrl->dr_version = 101;
	switch (mailbox) {
	case 0:
		{
			printf("cz%d: Cyclades 8Zo V1 Adapter. \n", czcontrol.active_boards);
			break;
		}
	case 1:
		{
			printf("cz%d: Cyclades 8Zo V2 Adapter. \n", czcontrol.active_boards);
			break;
		}
	case 2:
		{
			printf("cz%d: Cyclades Ze Adapter. \n", czcontrol.active_boards);
			break;
		}
	default:
		{
			printf("cz%d: Unknown Cyclades adapter. \n", czcontrol.active_boards);
			break;
		}
	}
	printf("cz%d: Driver Version %x. Firmware version %x. %d ports found.\n", czcontrol.active_boards, CZ_DRIVER_VERSION, czcontrol.board[czcontrol.active_boards].board_ctrl->fw_version, czcontrol.board[czcontrol.active_boards].board_ctrl->n_channel);
#ifndef CZINT
	printf("cz%d: Using POLLED mode. Trying to poll %d times per second.\n", czcontrol.active_boards, CZPOLLVAL);
	if ((hz / CZPOLLVAL) < 1)
		printf("cz%d: WARNING: Poll interval too high. Try a number like 25.\n", czcontrol.active_boards);
#else				/* CZINT */
	printf("cz%d: Using INTERRUPT mode.\n", czcontrol.active_boards);
#endif				/* CZINT */
	czcontrol.board[czcontrol.active_boards].number_of_ports = czcontrol.board[czcontrol.active_boards].board_ctrl->n_channel;
	czcontrol.board[czcontrol.active_boards].config_id = config_id;
	board = czcontrol.active_boards;

	for (i = 0; i < czcontrol.board[czcontrol.active_boards].board_ctrl->n_channel; i++) {
		bzero(&czcontrol.board[board].ports[i], sizeof(struct port_specific));
		while (czcontrol.board[board].plx9060->pci_doorbell != 0);
		czcontrol.board[czcontrol.active_boards].ports[i].ttyinfo.t_termios.c_ispeed = czcontrol.board[czcontrol.active_boards].ports[i].ttyinfo.t_termios.c_ospeed = 9600;
		termioschars(&czcontrol.board[czcontrol.active_boards].ports[i].it_in);
		termioschars(&czcontrol.board[czcontrol.active_boards].ports[i].it_out);
		termioschars(&czcontrol.board[czcontrol.active_boards].ports[i].ttyinfo.t_termios);
		termioschars(&czcontrol.board[czcontrol.active_boards].ports[i].ttyinfo.t_termios);

		czcontrol.board[czcontrol.active_boards].ports[i].it_in.c_cflag = TTYDEF_CFLAG | CRTSCTS;
		czcontrol.board[czcontrol.active_boards].ports[i].it_in.c_ispeed = czcontrol.board[czcontrol.active_boards].ports[i].it_in.c_ospeed = TTYDEF_SPEED;

		czcontrol.board[czcontrol.active_boards].ports[i].it_out.c_cflag = TTYDEF_CFLAG | CRTSCTS;
		czcontrol.board[czcontrol.active_boards].ports[i].it_out.c_ispeed = czcontrol.board[czcontrol.active_boards].ports[i].it_out.c_ospeed = TTYDEF_SPEED;

		czcontrol.board[czcontrol.active_boards].ports[i].channel_ctrl = (struct CH_CTRL *) & czcontrol.board[czcontrol.active_boards].zfw_ctrl->ch_ctrl[i];
		czcontrol.board[czcontrol.active_boards].ports[i].buffer_ctrl = (struct BUF_CTRL *) & czcontrol.board[czcontrol.active_boards].zfw_ctrl->buf_ctrl[i];
		czcontrol.board[czcontrol.active_boards].board_ctrl->hcmd_channel = i;
		czcontrol.board[czcontrol.active_boards].ports[i].channel_ctrl->op_mode = C_CH_ENABLE;
		czcontrol.board[czcontrol.active_boards].ports[i].channel_ctrl->intr_enable = C_CH_DISABLE;
		czcontrol.board[czcontrol.active_boards].ports[i].channel_ctrl->sw_flow = 0;
		czcontrol.board[czcontrol.active_boards].ports[i].channel_ctrl->comm_baud = 9600;
		czcontrol.board[czcontrol.active_boards].ports[i].channel_ctrl->comm_parity = C_PR_NONE;
		czcontrol.board[czcontrol.active_boards].ports[i].channel_ctrl->comm_data_l = C_DL_CS8 | C_DL_1STOP;
		czcontrol.board[czcontrol.active_boards].ports[i].channel_ctrl->hw_flow = 0;
		czcontrol.board[czcontrol.active_boards].ports[i].channel_ctrl->rs_control = 0;
		czcontrol.board[czcontrol.active_boards].plx9060->pci_doorbell = C_CM_IOCTLM;
		while (czcontrol.board[czcontrol.active_boards].plx9060->pci_doorbell != 0);
		czcontrol.board[czcontrol.active_boards].plx9060->pci_doorbell = C_CM_IOCTLM;
#ifdef DEVFS
		devfsminor = ((czcontrol.active_boards << 19) + i);
		sprintf(devfsname, "ttyC%c%02x\0", board_name[czcontrol.active_boards], i);
		devfs_token_ttyd = devfs_add_devswf(&cz_cdevsw, devfsminor, DV_CHR, UID_ROOT, GID_DIALER, 0660, devfsname);
		sprintf(devfsname, "ttyCi%c%02x\0", board_name[czcontrol.active_boards], i);
		devfs_token_ttyi = devfs_add_devswf(&cz_cdevsw, devfsminor + 131072, DV_CHR, UID_ROOT, GID_DIALER, 0660, devfsname);
		sprintf(devfsname, "ttyCl%c%02x\0", board_name[czcontrol.active_boards], i);
		devfs_token_ttyl = devfs_add_devswf(&cz_cdevsw, devfsminor + 262144, DV_CHR, UID_ROOT, GID_DIALER, 0660, devfsname);
		sprintf(devfsname, "cuaC%c%02x\0", board_name[czcontrol.active_boards], i);
		devfs_token_cuaa = devfs_add_devswf(&cz_cdevsw, devfsminor + 65536, DV_CHR, UID_ROOT, GID_DIALER, 0660, devfsname);
		sprintf(devfsname, "cuaCi%c%02x\0", board_name[czcontrol.active_boards], i);
		devfs_token_cual = devfs_add_devswf(&cz_cdevsw, devfsminor + 65536 + 131072, DV_CHR, UID_ROOT, GID_DIALER, 0660, devfsname);
		sprintf(devfsname, "cuaCl%c%02x\0", board_name[czcontrol.active_boards], i);
		devfs_token_cuai = devfs_add_devswf(&cz_cdevsw, devfsminor + 65536 + 262144, DV_CHR, UID_ROOT, GID_DIALER, 0660, devfsname);
#endif				/* DEVFS */

	}
	while (czcontrol.board[board].plx9060->pci_doorbell != 0);
	czcontrol.board[czcontrol.active_boards].plx9060->pci_doorbell = C_CM_IRQ_ENBL;
	while (czcontrol.board[board].plx9060->pci_doorbell != 0);

	if (czcontrol.active_boards == 0) {
		dev = makedev(CDEV_MAJOR, 0);
		cdevsw_add(&dev, &cz_cdevsw, NULL);
	}
#ifndef CZINT
	timeout(cz_poll, (void *) &czcontrol.board[czcontrol.active_boards], hz / CZPOLLVAL);
#endif				/* CZINT */
	czcontrol.active_boards++;
	return;
}

/* Open routine */
static int
cz_open(dev_t dev, int flag, int mode, struct proc * p)
{
	unsigned int    minordev, board, port, type, condition, sleep_return, spl_value = 0;
	struct tty     *tp;
	unsigned char   done = 0;
	minordev = minor(dev);
	board = CZBOARD(minordev);
	port = CZPORT(minordev);
	type = CZPORTTYPE(minordev);
	if (port > czcontrol.board[board].number_of_ports) {	/* Make sure they're not
								 * asking for a bogus
								 * port */
		return ENXIO;
	}
	if (type > 1)		/* Everyone can open initial and lock state,
				 * if access allows */
		return 0;

	spl_value = spltty();
	tp = &czcontrol.board[board].ports[port].ttyinfo;

open_top:

	while (czcontrol.board[board].ports[port].flags & DTR_FORCED_OFF) {
		sleep_return = tsleep(&czcontrol.board[board].ports[port].flags, TTIPRI | PCATCH, "czdtr", 0);
		if (sleep_return != 0) {
			splx(spl_value);
			return sleep_return;
		}
	}
	if (tp->t_state & TS_ISOPEN) {
		if (type == CALLOUT_PORT) {
			if ((czcontrol.board[board].ports[port].flags & PORT_CALL_OUT_ACTIVE) == 0) {
				splx(spl_value);
				return EBUSY;
			}
		} else {
			if (czcontrol.board[board].ports[port].flags & PORT_CALL_OUT_ACTIVE) {
				if (flag & O_NONBLOCK) {
					splx(spl_value);
					return EBUSY;
				}
				sleep_return = tsleep(&czcontrol.board[board].ports[port].flags, TTIPRI | PCATCH, "czbi", 0);
				if (sleep_return != 0) {
					splx(spl_value);
					return sleep_return;
				}
				goto open_top;
			}
		}
		if (tp->t_state & TS_XCLUDE && p->p_ucred->cr_uid != 0) {
			splx(spl_value);
			return EBUSY;
		}
	} else {
		if (czcontrol.board[board].ports[port].flags & LOCK_STATE_SET)
			tp->t_termios = (type == CALLOUT_PORT) ? czcontrol.board[board].ports[port].lt_out : czcontrol.board[board].ports[port].lt_in;
		else
			tp->t_termios = (type == CALLOUT_PORT) ? czcontrol.board[board].ports[port].it_out : czcontrol.board[board].ports[port].it_in;
		if ((czcontrol.board[board].ports[port].channel_ctrl->rs_status & C_RS_DCD) || (type == CALLOUT_PORT)) {
			(*linesw[tp->t_line].l_modem) (tp, 1);
		}
		ttsetwater(tp);
#ifdef CZINT
		czcontrol.board[board].ports[port].channel_ctrl->intr_enable = C_IN_TXBEMPTY | C_IN_TXLOWWM | C_IN_RXHIWM | C_IN_RXNNDT | C_IN_MDCD | C_IN_MDSR | C_IN_MRI | C_IN_MCTS;
#else				/* CZINT */
		czcontrol.board[board].ports[port].channel_ctrl->intr_enable = C_IN_DISABLE;
#endif				/* CZINT */
		czcontrol.board[board].ports[port].channel_ctrl->hw_flow = 0;
		czcontrol.board[board].ports[port].channel_ctrl->sw_flow = 0;
		czcontrol.board[board].ports[port].channel_ctrl->rs_control |=
			(C_RS_RTS | C_RS_DTR);
		cz_set_termios_params(board, port);

		if (!(tp->t_state & TS_CARR_ON) && !(type == CALLOUT_PORT) && !(tp->t_cflag & CLOCAL) && !(flag & O_NONBLOCK)) {
			++czcontrol.board[board].ports[port].openers;
			sleep_return = tsleep(TSA_CARR_ON(tp), TTIPRI | PCATCH, "czdcd", 0);
			if (sleep_return != 0) {
				splx(spl_value);
				--czcontrol.board[board].ports[port].openers;
				return sleep_return;
			}
			--czcontrol.board[board].ports[port].openers;
			goto open_top;
		}
		ttyopen(dev, &czcontrol.board[board].ports[port].ttyinfo);
		tp->t_oproc = cz_start;
		tp->t_param = cz_param;
		tp->t_dev = dev;
		czcontrol.board[board].ports[port].ttyinfo.t_state &= ~(TS_BUSY);
		sleep_return = (*linesw[tp->t_line].l_open) (dev, tp);
		disc_optim(tp, &tp->t_termios, &czcontrol.board[board].ports[port]);
		if (tp->t_state & TS_ISOPEN && type == CALLOUT_PORT)
			czcontrol.board[board].ports[port].flags |= PORT_CALL_OUT_ACTIVE;
	}
	tp->t_state |= TS_ISOPEN;
	splx(spl_value);
	return (0);
}

/* Read routine (just passes control to the line disc.) */
static int
cz_read(dev_t dev, struct uio * uio, int flag)
{
	int             i;
	unsigned int    minordev, board, port, type = 0;
	struct tty     *tp;
	minordev = minor(dev);
	board = CZBOARD(minordev);
	port = CZPORT(minordev);
	type = CZPORTTYPE(minordev);
	if (type > 1)
		return ENXIO;
	tp = &czcontrol.board[board].ports[port].ttyinfo;
	return ((*linesw[tp->t_line].l_read) (tp, uio, flag));
}

/* Write routine (just passes control to the line disc.) */
static int
cz_write(dev_t dev, struct uio * uio, int flag)
{
	unsigned int    minordev, board, port, type = 0;
	int             return_val;
	struct tty     *tp;
	minordev = minor(dev);
	board = CZBOARD(minordev);
	port = CZPORT(minordev);
	type = CZPORTTYPE(minordev);
	tp = &czcontrol.board[board].ports[port].ttyinfo;
	if (type > 1)
		return ENXIO;
	return (*linesw[tp->t_line].l_write) (tp, uio, flag);
}

/* IOCTL Routine */
static int
cz_ioctl(dev_t dev, int cmd, caddr_t data, int flag, struct proc * p)
{
	int             error = 0;
	int             minordev, board, port, type = 0;
	int             spl_value = 0;
	struct tty     *tp;
	struct port_specific *mtp;
	struct termios *debug_dt;
	struct termios *dt = (struct termios *) data;

#if defined(COMPAT_43) || defined(COMPAT_SUNOS)
	int             oldcmd;
	struct termios  term;
#endif
	minordev = minor(dev);
	board = CZBOARD(minordev);
	port = CZPORT(minordev);
	type = CZPORTTYPE(minordev);
	mtp = &czcontrol.board[board].ports[port];

	tp = &czcontrol.board[board].ports[port].ttyinfo;

#if defined(COMPAT_43) || defined(COMPAT_SUNOS)
	term = tp->t_termios;
	oldcmd = cmd;
	error = ttsetcompat(tp, &cmd, data, &term);
	if (error != 0)
		return error;
	if (cmd != oldcmd)
		data = (caddr_t) & term;
#endif

	if (type > 1) {
		struct termios *ct;
		switch (type) {
		case 2:
		case 3:
			{
				ct = type & CALLOUT_INITIAL ? &mtp->it_out : &mtp->it_in;
				break;
			}
		case 4:
		case 5:
			{
				ct = type & CALLOUT_LOCK ? &mtp->lt_out : &mtp->lt_in;
				break;
			}
		default:
			{
				return ENODEV;
				break;
			}
			switch (cmd) {
			case TIOCSETA:
				{
					error = suser(p->p_ucred, &p->p_acflag);
					if (error != 0)
						return error;
					*ct = *(struct termios *) data;
					return 0;
					break;
				}
			case TIOCGETA:
				{
					*(struct termios *) data = *ct;
					return 0;
					break;
				}
			case TIOCGETD:
				{
					*(int *) data = TTYDISC;
					return 0;
					break;
				}
			case TIOCGWINSZ:
				{
					bzero(data, sizeof(struct winsize));
					return 0;
					break;
				}
			default:
				{
					return ENOTTY;
					break;
				}
			}
		}
	}
	if (cmd == TIOCSETA || cmd == TIOCSETAW || cmd == TIOCSETAF) {
		int             cc;
		struct termios *lt = (type & CALLOUT_LOCK) ? &mtp->lt_out : &mtp->lt_in;

		dt->c_iflag = (tp->t_iflag & lt->c_iflag) | (dt->c_iflag & ~lt->c_iflag);

		dt->c_oflag = (tp->t_oflag & lt->c_oflag) | (dt->c_oflag & ~lt->c_oflag);

		dt->c_cflag = (tp->t_cflag & lt->c_cflag) | (dt->c_cflag & ~lt->c_cflag);

		dt->c_lflag = (tp->t_lflag & lt->c_lflag) | (dt->c_lflag & ~lt->c_lflag);

		for (cc = 0; cc < NCCS; ++cc)
			if (lt->c_cc[cc] != 0)
				dt->c_cc[cc] = tp->t_cc[cc];

		if (lt->c_ispeed != 0)
			dt->c_ispeed = tp->t_ispeed;
		if (lt->c_ospeed != 0)
			dt->c_ospeed = tp->t_ospeed;
	}
	error = 0;
	error = (*linesw[tp->t_line].l_ioctl) (tp, cmd, data, flag, p);
	if (error >= 0) {
		return error;
	}
	spl_value = spltty();
	error = ttioctl(tp, cmd, data, flag);
	disc_optim(tp, &tp->t_termios, &czcontrol.board[board].ports[port]);
	if (error >= 0) {
		splx(spl_value);
		return error;
	}
	switch (cmd) {
	case TIOCSBRK:
		{
			while (czcontrol.board[board].plx9060->pci_doorbell != 0);
			czcontrol.board[board].board_ctrl->hcmd_channel = port;
			czcontrol.board[board].plx9060->pci_doorbell = C_CM_SET_BREAK;
			break;
		}
	case TIOCCBRK:
		{
			while (czcontrol.board[board].plx9060->pci_doorbell != 0);
			czcontrol.board[board].board_ctrl->hcmd_channel = port;
			czcontrol.board[board].plx9060->pci_doorbell = C_CM_CLR_BREAK;
			break;
		}
	case TIOCSDTR:
		{
			while (czcontrol.board[board].plx9060->pci_doorbell != 0);
			czcontrol.board[board].board_ctrl->hcmd_channel = port;
			czcontrol.board[board].ports[port].channel_ctrl->rs_control |= C_RS_DTR;
			czcontrol.board[board].plx9060->pci_doorbell = C_CM_IOCTLM;
			break;
		}
	case TIOCCDTR:
		{
			while (czcontrol.board[board].plx9060->pci_doorbell != 0);
			czcontrol.board[board].board_ctrl->hcmd_channel = port;
			czcontrol.board[board].ports[port].channel_ctrl->rs_control &= ~(C_RS_DTR);
			czcontrol.board[board].plx9060->pci_doorbell = C_CM_IOCTLM;
			break;
		}
	case TIOCMSET:
		{
			while (czcontrol.board[board].plx9060->pci_doorbell != 0);
			if (*data & TIOCM_RTS)
				czcontrol.board[board].ports[port].channel_ctrl->rs_control |= C_RS_RTS;
			else
				czcontrol.board[board].ports[port].channel_ctrl->rs_control &= ~(C_RS_RTS);
			if (*data & TIOCM_CTS)
				czcontrol.board[board].ports[port].channel_ctrl->rs_control |= C_RS_CTS;
			else
				czcontrol.board[board].ports[port].channel_ctrl->rs_control &= ~(C_RS_CTS);

			if (*data & TIOCM_LE)
				czcontrol.board[board].ports[port].channel_ctrl->op_mode = C_CH_ENABLE;
			else
				czcontrol.board[board].ports[port].channel_ctrl->op_mode = C_CH_DISABLE;
			if (*data & TIOCM_DTR)
				czcontrol.board[board].ports[port].channel_ctrl->rs_control |= C_RS_DTR;
			else
				czcontrol.board[board].ports[port].channel_ctrl->rs_control &= ~(C_RS_DTR);

			if (*data & TIOCM_DSR)
				czcontrol.board[board].ports[port].channel_ctrl->rs_control |= C_RS_DSR;
			else
				czcontrol.board[board].ports[port].channel_ctrl->rs_control &= ~(C_RS_DSR);

			czcontrol.board[board].board_ctrl->hcmd_channel = port;
			czcontrol.board[board].plx9060->pci_doorbell = C_CM_IOCTLM;
			break;
		}
	case TIOCMGET:
		{
			*data = 0;
			if (czcontrol.board[board].ports[port].channel_ctrl->rs_status & C_RS_RTS)
				*data |= TIOCM_RTS;
			if (czcontrol.board[board].ports[port].channel_ctrl->rs_status & C_RS_CTS)
				*data |= TIOCM_CTS;
			if (czcontrol.board[board].ports[port].channel_ctrl->rs_status & C_RS_DCD)
				*data |= TIOCM_CAR;

			if (czcontrol.board[board].ports[port].channel_ctrl->op_mode & C_CH_ENABLE)
				*data |= TIOCM_LE;
			if (czcontrol.board[board].ports[port].channel_ctrl->rs_status & C_RS_DTR)
				*data |= TIOCM_DTR;
			if (czcontrol.board[board].ports[port].channel_ctrl->rs_status & C_RS_RI)
				*data |= TIOCM_RNG;
			if (czcontrol.board[board].ports[port].channel_ctrl->rs_status & C_RS_DSR)
				*data |= TIOCM_DSR;
			break;
		}
	case TIOCMBIS:
		{
			while (czcontrol.board[board].plx9060->pci_doorbell != 0);
			if (*data & TIOCM_RTS)
				czcontrol.board[board].ports[port].channel_ctrl->rs_control |= C_RS_RTS;
			if (*data & TIOCM_CTS)
				czcontrol.board[board].ports[port].channel_ctrl->rs_control |= C_RS_CTS;
			if (*data & TIOCM_CAR)
				czcontrol.board[board].ports[port].ttyinfo.t_termios.c_cflag |= CLOCAL;
			if (*data & TIOCM_LE)
				czcontrol.board[board].ports[port].channel_ctrl->op_mode = C_CH_ENABLE;
			if (*data & TIOCM_DTR)
				czcontrol.board[board].ports[port].channel_ctrl->rs_control |= C_RS_DTR;
			if (*data & TIOCM_DSR)
				czcontrol.board[board].ports[port].channel_ctrl->rs_control |= C_RS_DSR;
			czcontrol.board[board].board_ctrl->hcmd_channel = port;
			czcontrol.board[board].plx9060->pci_doorbell = C_CM_IOCTLM;
			break;
		}
	case TIOCMBIC:
		{
			while (czcontrol.board[board].plx9060->pci_doorbell != 0);
			if (*data & TIOCM_RTS)
				czcontrol.board[board].ports[port].channel_ctrl->rs_control &= ~(C_RS_RTS);
			if (*data & TIOCM_CTS)
				czcontrol.board[board].ports[port].channel_ctrl->rs_control &= ~(C_RS_CTS);
			if (*data & TIOCM_CAR)
				czcontrol.board[board].ports[port].ttyinfo.t_termios.c_cflag &= ~(CLOCAL);
			if (*data & TIOCM_LE)
				czcontrol.board[board].ports[port].channel_ctrl->op_mode = C_CH_DISABLE;
			if (*data & TIOCM_DTR)
				czcontrol.board[board].ports[port].channel_ctrl->rs_control &= ~(C_RS_DTR);

			if (*data & TIOCM_DSR)
				czcontrol.board[board].ports[port].channel_ctrl->rs_control &= ~(C_RS_DSR);
			czcontrol.board[board].board_ctrl->hcmd_channel = port;
			czcontrol.board[board].plx9060->pci_doorbell = C_CM_IOCTLM;
			break;
		}
	case TIOCMSDTRWAIT:
		{
			error = suser(p->p_ucred, &p->p_acflag);
			if (error != 0) {
				splx(spl_value);
				return (error);
			}
			czcontrol.board[board].ports[port].dtrwait = *(int *) data *hz / 100;
			break;
		}
	case TIOCMGDTRWAIT:
		{
			*(int *) data = czcontrol.board[board].ports[port].dtrwait * 100 / hz;
			break;
		}
	default:
		{
			splx(spl_value);
			return ENOTTY;
		}
	}
	/*
	 * We shouldn't need these at all, as ioctl()'s should call cz_param,
	 * or something similar to do the work for us. However, if you need
	 * think you need it, uncomment it
	 */
	/*
	 * disc_optim(tp, &tp->t_termios,
	 * &czcontrol.board[board].ports[port]); cz_set_termios_params(board,
	 * port);
	 */

	splx(spl_value);
	return 0;
}

/* Close routine */
static int
cz_close(dev_t dev, int flag, int mode, struct proc * p)
{
	unsigned int    minordev, board, port, type, condition, sleep_return, spl_value = 0;
	struct tty     *tp;
	minordev = minor(dev);
	board = CZBOARD(minordev);
	port = CZPORT(minordev);
	type = CZPORTTYPE(minordev);
	if (port > czcontrol.board[board].number_of_ports)	/* Make sure they're not
								 * asking for a bogus
								 * port */
		return ENXIO;

	if (type > 1)
		return 0;

	tp = &czcontrol.board[board].ports[port].ttyinfo;
	spl_value = spltty();
	(*linesw[tp->t_line].l_close) (tp, flag);

	disc_optim(tp, &tp->t_termios, &czcontrol.board[board].ports[port]);

	while (czcontrol.board[board].plx9060->pci_doorbell != 0);
	czcontrol.board[board].ports[port].channel_ctrl->intr_enable = 0;
	czcontrol.board[board].board_ctrl->hcmd_channel = port;
	czcontrol.board[board].plx9060->pci_doorbell = C_CM_IOCTLW;
	while (czcontrol.board[board].plx9060->pci_doorbell != 0);
	czcontrol.board[board].ports[port].channel_ctrl->rs_control &=
		~(C_RS_RTS | C_RS_DTR);
	czcontrol.board[board].ports[port].flags |= DTR_FORCED_OFF;
	czcontrol.board[board].board_ctrl->hcmd_channel = port;
	czcontrol.board[board].plx9060->pci_doorbell = C_CM_IOCTLM;
	cz_stop(tp, FREAD | FWRITE);
	ttyclose(tp);
	timeout(clear_dtr_flag, (void *) &czcontrol.board[board].ports[port].ttyinfo.t_dev, CZ_DTR_TICKS);
	splx(spl_value);
	return 0;
}


/* Line discp. optimization routine */
static void
disc_optim(tp, t, com)
	struct tty     *tp;
	struct termios *t;
	struct port_specific *com;
{
	if (!(t->c_iflag & (ICRNL | IGNCR | IMAXBEL | INLCR | ISTRIP | IXON))
	    && (!(t->c_iflag & BRKINT) || (t->c_iflag & IGNBRK))
	    && (!(t->c_iflag & PARMRK)
		|| (t->c_iflag & (IGNPAR | IGNBRK)) == (IGNPAR | IGNBRK))
	    && !(t->c_lflag & (ECHO | ICANON | IEXTEN | ISIG | PENDIN))
	    && linesw[tp->t_line].l_rint == ttyinput)
		tp->t_state |= TS_CAN_BYPASS_L_RINT;
	else
		tp->t_state &= ~TS_CAN_BYPASS_L_RINT;
	/*
	 * Prepare to reduce input latency for packet discplines with a end
	 * of packet character.
	 */
	if (tp->t_line == SLIPDISC)
		com->hotchar = 0xc0;
	else if (tp->t_line == PPPDISC)
		com->hotchar = 0x7e;
	else
		com->hotchar = 0;
}

/* Dev to tty routine for line disc. */
static struct tty *
cz_devtotty(dev_t dev)
{
	int             minordev, board, port, type;
	minordev = minor(dev);
	board = CZBOARD(minordev);
	port = CZPORT(minordev);
	type = CZPORTTYPE(minordev);
	if (type > 1)
		return NULL;
	if (port >= czcontrol.board[board].number_of_ports)
		return NULL;
	return (&czcontrol.board[board].ports[port].ttyinfo);
}

/*
 * Routine for calculating how many bytes we can transmit in one memory copy
 * operation. This really should be replaced by macros, instead, but I've had
 * some problems with macros returning weird numbers, and there are places
 * where it may generate recursive macros, so, we have a function, for
 * now....
 */

unsigned long int
tx_moveable(unsigned long int put, unsigned long int get, unsigned long int size)
{
	if (put == get)
		return (size - put);
	if (put > get)
		return (size - put);
	return ((get - put) - 1);
}

/*
 * Routine for calculating how many bytes we can receive in one memory copy
 * operation. This really should be replaced by macros, instead, but I've had
 * some problems with macros returning weird numbers, and there are places
 * where it may generate recursive macros, so, we have a function, for
 * now....
 */

unsigned long int
rx_moveable(unsigned long int put, unsigned long int get, unsigned long int size)
{
	if (put == get) {
		return 0;
	}
	if (put > get) {
		return (put - get);
	}
	return (size - get);
}

#endif				/* NCZ */
